#ifndef __XEN_X86_VENDORS_H__
#define __XEN_X86_VENDORS_H__

/*
 * CPU vendor IDs
 */
#define X86_VENDOR_INTEL 0
#define X86_VENDOR_AMD 1
#define X86_VENDOR_CENTAUR 2
#define X86_VENDOR_NUM 3
#define X86_VENDOR_UNKNOWN 0xff

#endif	/* __XEN_X86_VENDORS_H__ */
